| Routines for FFT calculation with floating points values
|
| Original Amiga file by Stephane TAVENARD
|
| 68030/68882 code optimized on Falcon by Draco:
| - main loop: 	removed unnecessary stack push/pop operations
| - energy_phi: removed unnecessary ftst.x instruction
|		removed unnecessary bra.s (one or two)
|		removed unnecessary fmove.s #0,fp1 (ever heard of fmovecr, BTW?!)
| - both:	instruction reordering to gain some speed from 68882 CU/APU concurrency
|		added wrapper for gcc passing arguments
|
|
	.globl	   _ASM_FFT_main_loop
	.globl	   _ASM_FFT_energy_phi
|
|		  even

|		  a0: float *x_real
|		  a1: float *x_imag
|		  a2: float *cos
|		  a3: float *sin
|		  d0: power
|
_ASM_FFT_main_loop:
	moveml	d0-d6/a0-a5,sp@-
	movel	sp@(56),a0
	movel	sp@(60),a1
	movel	sp@(64),a2
	movel	sp@(68),a3
	movel	sp@(72),d0
	movel	a2,a4	  		| a4 = cos
	movel	a3,a5	  		| a5 = sin
	moveql	#1,d1	  		| d1 = d = 1
	clrl	d2
	bset	d0,d2
	asrl	#1,d2	  		| d2 = li = 1<<( power-1 )
	subqw	#1,d0	  		| d0 = i
	movel	a0,d5
	movel	a1,d6
FFT_main_loop_1:
	lea	a0@(0,d1:w:4),a2	| a2 = &x_real[ d ]
	lea	a1@(0,d1:w:4),a3	| a3 = &x_imag[ d ]
	movew	d2,d3
	subqw	#1,d3	  		| d3 = j
FFT_main_loop_2:
	movew	d1,d4
	subqw	#1,d4	  		| d4 = l
FFT_main_loop_3:
	fmoves	a0@,fp0	  		| fp0 = *real_ptr
	fmoves	a2@,fp2			| fp2 = *t_real_ptr
	fmoves	a1@,fp1			| fp1 = *imag_ptr
	fmoves	a3@,fp3			| fp3 = *t_imag_ptr
	fmovex	fp2,fp4
	faddx	fp0,fp2	  		| fp2 = *real_ptr + *t_real_ptr
	fmovex	fp3,fp5
	faddx	fp1,fp3	  		| fp3 = *imag_ptr + *t_imag_ptr
	fmoves	fp2,a0@+	  	| fp2 -> *real_ptr++
	fsubx	fp4,fp0	  		| fp0 = *real_ptr - *t_real_ptr
	fmoves	fp3,a1@+	  	| fp3 -> *imag_ptr++
	fsubx	fp5,fp1	  		| fp1 = *imag_ptr - *t_imag_ptr
	fmoves	fp0,a2@+	  	| fp4 -> *t_real_ptr++
	fmoves	fp1,a3@+	  	| fp5 -> *t_imag_ptr++
	dbra	d4,FFT_main_loop_3
	lea	a0@(0,d1:w:4),a0 	| real_ptr += d;
	lea	a1@(0,d1:w:4),a1 	| imag_ptr += d;
	lea	a2@(0,d1:w:4),a2 	| t_real_ptr += d;
	lea	a3@(0,d1:w:4),a3 	| t_imag_ptr += d;
	dbra	d3,FFT_main_loop_2

	movel	d5,a0			| restore a0/a1 (was pop/push here)
	movel	d6,a1

	asll	#1,d1			| d = d<<1;
	asrl	#1,d2			| li = li>>1;
	beqs	FFT_main_loop_9		| li = 0 -> end of FFT
	lea	a0@(0,d1:w:4),a0	| a0 = real_ptr = &x_real[ d ]
	lea	a1@(0,d1:w:4),a1	| a1 = imag_ptr = &x_imag[ d ]
	movew	d2,d3
	subqw	#1,d3			| d3 = j
FFT_main_loop_5:
	movel	a4,a2			| a2 = cos_ptr = cos
	movel	a5,a3			| a3 = sin_ptr = sin
	movew	d1,d4
	subqw	#1,d4			| d4 = l
FFT_main_loop_6:
	fmoves  a0@,fp0			| fp0 = *real_ptr
	fmoves  a1@,fp1			| fp1 = *imag_ptr
	fmoves  a2@,fp2			| fp2 = *cos_ptr
	fmoves  a3@,fp3			| fp3 = *sin_ptr
	fmovex  fp2,fp4
	fsglmulx fp0,fp2		| fp2 = (*real_ptr)*(*cos_ptr)
	fmovex  fp3,fp5
	fsglmulx fp1,fp3		| fp3 = (*imag_ptr)*(*sin_ptr)
	fsglmulx fp1,fp4		| fp4 = (*imag_ptr)*(*cos_ptr)
	faddx   fp2,fp3	  		| fp3 = real
	fsglmulx fp0,fp5		| fp5 = (*real_ptr)*(*sin_ptr)
	fsubx	fp5,fp4			| fp4 = imag
	fmoves	fp3,a0@+		| *real_ptr++ = real
	fmoves	fp4,a1@+		| *imag_ptr++ = imag
	lea	a2@(0,d2:w:4),a2	| cos_ptr += li
	lea	a3@(0,d2:w:4),a3	| sin_ptr += li
	dbra	d4,FFT_main_loop_6
	lea	a0@(0,d1:w:4),a0	| real_ptr += d
	lea	a1@(0,d1:w:4),a1	| imag_ptr += d
	dbra	d3,FFT_main_loop_5
FFT_main_loop_9:
	movel	d5,a0
	movel	d6,a1
	dbra	d0,FFT_main_loop_1
	moveml	sp@+,d0-d6/a0-a5
	rts

|		  even

|		  a0: float *x_real
|		  a1: float *x_imag
|		  a2: float *energy
|		  a3: float *phi
|		  d0: n
|
_ASM_FFT_energy_phi:
	moveml	d0/a0-a3,sp@-
	movel	sp@(24),a0
	movel	sp@(28),a1
	movel	sp@(32),a2
	movel	sp@(36),a3
	movel	sp@(40),d0	
	subqw   #1,d0
FFT_phi0:
	fmoves  a0@+,fp0
	fmoves  a1@+,fp1
	fmovex  fp0,fp4
	fmovex  fp1,fp5
	fsglmulx fp4,fp4
	fsglmulx fp5,fp5
	faddx   fp4,fp5			| fp5 = en = real^2 + imag^2
	fmoves  fp5,a2@+	     	| *energy++ = en
	fbeq	FFT_phi3	  	| en <> 0 (flags reflect the result of faddx)
FFT_phi1:
	ftstx   fp0
	fbne	FFT_phi2	  	| real == 0
	fmoves	#1/1000000000,fp0	| 1E-20
FFT_phi2:
	fsgldivx fp0,fp1		| fp1 = imag / real
	fatanx	fp1,fp1			| fp1 = atan( imag / real )
FFT_phi5:
	fmoves  fp1,a3@+		| *phi++ = phi
	dbra	d0,FFT_phi0
	bras	FFT_exit
FFT_phi3:
	clrl	a3@+			| *phi++ = 0
	dbra	d0,FFT_phi0
FFT_exit:
	moveml	sp@+,d0/a0-a3
	rts
